{{
/**
 * @file jazzedmain.hh
 *
  Jazzed_JVM_Main_002
    * Found bug in heap code. heap pointers should be word aligned.
      This was causing strange print results like missing cr/lf in
      System.out.println("Hello World");
    * Changed register dump to listing for c,f,m,n,s commands.
      V verbosity still is registerdump.
    * Added end of main detection so program stops on main exit.
    * Removed DEAD from heap dump; not a valid message for type code.
      
    added new features
        R: show registers  
    changed behaviour in features
        j: show JVM regs & stack - was show JVM status. heap dump too big. 
        l: list bytecode assembly - list 10 instructions after current PC line. was variable amount.
        n: next instruction in method - now prints two line disassembly after run.
        s: single step instruction  - now prints two line disassembly after run.
    fixed bugs in features
        n: next instruction in method - this could crash on return instructions. fixed.
        o: dump an object contents - this could crash if ptr not an object. fixed. 
        p: print memory range - entering p again should ask for new address. hit return continues old dump. 
        
  Jazzed_JVM_Main_001
    initial release with
        b: set breakpoint
        c: continue debugging
        C: clear all breakpoints
        d: delete a given breakpoint
        f: finish a method
        h: dump heap
        i: init JVM to reset state
        j: show JVM status. 
        l: list bytecode assembly
        L: list all breakpoints
        m: break on new method
        n: next instruction in method
        o: dump an object contents 
        p: print memory range 
        q: quit and reboot
        r: run to a breakpoint  
        s: single step instruction
        S: dump stack
        v: set callDebug verbosity 
        V: set more detail verbosity 
        u: unset verbosity 
        x: examine one memory address 
        Enter repeats last command.
 */
}}

CON
' .-----------------------------------------------------------------------------------------------------.
' |     Board Configuration                                                                             |
' `-----------------------------------------------------------------------------------------------------'
  _CLKMODE            = XTAL1 + PLL16x                  ' 80MHz System Clock Speed for prop protoboard
  _XINFREQ            = 5_000_000

CON

  SIO_SPEED = 28800
  SBUFLEN = 128
  BREAKPOINT_SIZE = 10
  
OBJ

  sio: "Jazzed_SIO_Text"                                ' Serial IO
  jvm: "Jazzed_Jvm_Engine_010"                          ' JVM engine
  ops: "Jazzed_JVM_OpStrings_001"                       ' JVM debug opcode strings
  prg: "Jazzed_JVM_JemFile_003"                         ' The jem file for JVM to run
  jem: "Jazzed_JVM_JemCodes_001"                        ' jemcodes file
  jni: "Jazzed_JVM_jvmNative"                           ' native functions
  
VAR

  word gGetOverBreak                                    ' skip break once if set during isBreakPoint
  word gBreakpointHead                                  ' breakpoint array queue head
  word gBreakpointArray[BREAKPOINT_SIZE]                ' breakpoint array

PUB Main | last, ofp, omp, opc, sbuf[SBUFLEN+1], printing, debug
{{              
/**
 * Main entry point for standalone jvm testing.
 */
__ void main(); 
}}
  sio.start(31, 30, 0, SIO_SPEED)
  
  jvm.start
  jvm.setJvmStart(prg.ByteCodeStart)

  if getMaxJvmBreakpoints <> BREAKPOINT_SIZE
    sio.str(string($d,"Code err: Breakpoint size <> JVM."))
    repeat
    
  clearAllJvmBreakpoints
  printMenu
  
  jvm.jvmInit
  printing := 0

  repeat
    sio.Str(String($d,"JVM> "))
    result := sio.getch
    if result <> $D
      last := result
      ' if last is p and we're printing, get a new address.
      if last == "p" and printing
        printing := 0
    sio.out(last)
    sio.out(" ")
    case last
      "b": ' set breakpoint
        sio.Str(String(" JemPC $> "))
        result := getBreakAddr
        if result <> 0
          'sio.hex(result,4)
          setJvmBreakpoint(result+jvm.getJvmStart)
        result := 0
      "c": ' continue debugging
        jvm.skipJvmBreakPointOnce 
        repeat
          if sio.key
            quit
          if jvmDebug(1,debug) < 0
            quit
        jvmDumpBreak
      "C": ' clear all breakpoints
        clearAllJvmBreakpoints
      "d": ' delete a given breakpoint
        sio.Str(String("BP #> "))
        'result := getBreakAddr
        result := getBrkPtNum
        if result <> 0
          'sio.hex(result,4)
          'deleteJvmBreakpoint(result+jvm.getJvmStart)
          deleteJvmBrkPtByNum(result)
      "f": ' singlestep until frame changes
        jvm.skipJvmBreakPointOnce
        ofp := jvm.getJvmFP
        repeat
          if jvmDebug(1,debug) < 0
            quit
          if sio.key <> 0
            sio.getch
            quit
        while (jvm.getJvmFP == ofp)
        'jvmDumpBreak
        asmBytecode(0, 2)
      "g": ' collect garbage
        CleanHeap
      "h": ' dump heap
        ShowHeap
      "i": ' reinit jvm
        jvm.jvmInit
        jvm.resetJvmPc
      "j": ' jvm status
        dumpRegs
        showStack
      "l": ' list disassembly of bytecodes from PC-2 instructions if possible to PC+20
        asmBytecode(0, 10)
      "L": ' list all breakpoints
        listJvmBreakpoints
      "m": ' singlestep until method changes
        omp := jvm.getJvmMP
        jvm.skipJvmBreakPointOnce
        repeat
          if jvmDebug(1,debug) < 0
            quit
          if sio.key <> 0
            sio.getch
            quit
        while (jvm.getJvmMP == omp)
        'jvmDumpBreak
        asmBytecode(0, 2)
      "n": ' singlestep over function
        omp := jvm.getJvmMP
        opc := jvm.getJvmPC
        jvm.skipJvmBreakPointOnce
        repeat
          jvmDebug(1,debug)
          if sio.key <> 0
            sio.getch
            quit
          if byte[opc] == jem#Jem_RETURN
            quit
          if byte[opc] == jem#Jem_IRETURN
            quit
        while (jvm.getJvmMP <> omp)
        asmBytecode(0, 2)
      "o": ' dump an object
        sio.Str(String(" ADDR $> "))
        sio.fgets(@sbuf, SBUFLEN)
        result := sio.strtohex(@sbuf)
        ShowObject(result)      
      "p": ' print memory
        if printing == 0
          sio.Str(String(" ADDR $> "))
          sio.fgets(@sbuf, SBUFLEN)
          result := sio.strtohex(@sbuf)
          dumpBytes(result,$80)
        else      
          dumpBytes(printing,$80)
        printing := result+$80
      "q": ' reboot
        reboot
      "r": ' run to a breakpoint
        jvm.jvmInit
        jvm.resetJvmPc
        sio.out($d)   
        repeat                                                                                                                                                                 
          if sio.key
            sio.getch
            quit
          if jvmDebug(1,debug) < 0
            quit
        jvmDumpBreak
      "R": ' dump registers
        jvmDumpBreak
      "s": ' single step into function
        jvm.skipJvmBreakPointOnce
        jvmDebug(1,debug)
        asmBytecode(0, 2)
      "S": ' dump stack
        ShowStack
      "v": ' set callDebug verbosity
        jvm.setJvmCallDebug(1) 
      "V": ' set instruction level verbosity
        debug := 1 
        jvm.setJvmVerboseDebug(1)  
      "u": ' unset verbosity                                                                                                                        
        jvm.setJvmCallDebug(0)
        debug := 0 
        jvm.setJvmVerboseDebug(0)
      "x": ' x-amine memory location hex value
        sio.Str(String(" ADDR $> "))
        sio.fgets(@sbuf, SBUFLEN)
        result := sio.strtohex(@sbuf)
        sio.Str(string("ADDR "))
        sio.hex(result, 4)
        sio.Str(string(" = $"))
        sio.hex(word[result], 2)
      " ":
      other:
        printMenu
  
PRI printMenu
  sio.cls 
  sio.out($d)
  sio.out($d)
  sio.str(string($d,"JVM Human Interface Debug Help:"))
  sio.str(string($d," b: set breakpoint"))
  sio.str(string($d," c: continue debugging"))
  sio.str(string($d," C: clear all breakpoints"))
  sio.str(string($d," d: delete a given breakpoint"))
  sio.str(string($d," f: finish a method"))
  sio.str(string($d," h: dump heap"))
  sio.str(string($d," i: init JVM to reset state"))
  sio.str(string($d," j: show JVM regs & stack")) 
  sio.str(string($d," l: list bytecode assembly"))
  sio.str(string($d," L: list all breakpoints"))
  sio.str(string($d," m: break on new method"))
  sio.str(string($d," n: next instruction in method"))
  sio.str(string($d," o: dump an object contents")) 
  sio.str(string($d," p: print memory range")) 
  sio.str(string($d," q: quit and reboot"))
  sio.str(string($d," r: run to a breakpoint"))  
  sio.str(string($d," R: show registers"))  
  sio.str(string($d," s: single step instruction"))
  sio.str(string($d," S: dump stack"))
  sio.str(string($d," v: set callDebug verbosity")) 
  sio.str(string($d," V: set more detail verbosity")) 
  sio.str(string($d," u: unset verbosity")) 
  sio.str(string($d," x: examine one memory address")) 
  sio.str(string($d," Enter repeats last command."))
  'sio.str(string($d," Use 'n' to skip breakpoint."))
  sio.out($d)
  

PUB jvmDebug(stp, verb) | pc, opcode

  pc := jvm.getJvmPC
  opcode := byte[pc]

  if jvm.isBreakpoint(pc)
    return -1
    
  result := jvm.jvmSpinDebug(stp, 0, ops.OpcodeNamePtr(opcode))
  
  handleNative(result)
      
  if verb
    jvmDumpBreak

  if jvm.getJvmMP == 0 or jvm.getJvmFP == 0
    return -1
    
  result := 0


PRI handleNative(code) | stk, ii, rcnt, val

  ' native methods are positive integers
  if code < 0
    return
{    
  sio.str(string($d,"Run Native "))
  sio.hex(code,2)
}
  rcnt := jvm.Pop
{
  sio.str(string(" P# "))
  sio.dec(rcnt)
  sio.out(" ")
}
  repeat while (rcnt-- > 0)
    val := jvm.GetVar(rcnt)
    'sio.hex(val,4)
    'sio.out(" ")
    result := jvm.GetHeapValuePtr(val,0)
    if result <> 0
      val := result
    jni.push(val)

  result := jni.do_native(code, @rcnt)
{
  sio.str(string(" R# "))
  sio.dec(rcnt)
}
  if rcnt > 0
    jvm.Push(result)
   
PUB jvmDumpBreak | pc, opcode
 pc := jvm.getJvmPC
  opcode := byte[pc]
  if jvm.isBreakpoint(pc)
    sio.str(string($d,"Breakpoint PC $"))
    sio.hex(pc,4)
    sio.str(string(" JemPC $"))
    sio.hex(pc-jvm.getJvmStart,4)
    sio.out($d)
  JvmSpinDumpBreak(opcode, ops.OpcodeNamePtr(opcode))

PUB asmBytecode(start, count) | ii, pc, mb, opc, bpt, lines
  pc := jvm.getJvmPC
  mb := jvm.getJvmMP
  ii := mb-1
  if start == 0
    ii := pc-1
  lines := count
  sio.out($d)
  repeat 'while(++ii < pc+count)
    ++ii
    if(--lines < 0)
      quit
    opc := byte[ii]
    bpt := jvm.isBreakpoint(ii)
    sio.str(string($d,"JPC "))
    if bpt
      sio.hex(bpt,1)
    else
      sio.out("-")
    sio.out(" ")
    sio.out("$")
    sio.hex(ii-jvm.getJvmStart,4)
    sio.out(" ")
    if(ii == pc)
      sio.out("*")
    else
      sio.out(" ")
    dumpOpcode(opc, ops.OpcodeNamePtr(opc))
    ii += ops.getOperandSize(opc) ' skip opcodes
  sio.out($d) 
    
PUB dumpBytes(addr,len) | ndx, bytes[17], bi, sz
{{
/**
 * dump bytes in 8 column table
 * @param addr is base address for dump
 * @param len is number of bytes to dump
 */
__ void dumpBytes(addr,len);
}}
  ndx := 0
  sz  := 16
  'str(string("Dump"))
  repeat while(ndx < len)
    if ndx <> 0 and (ndx // sz) <> 0
      sio.out(".")
      bytes[bi++ // sz] := byte[addr+ndx] 
    if ndx == 0 or (ndx // sz) == 0
      if ndx <> 0
        sio.out(" ")
        repeat bi from 0 to sz-1
          if (bytes[bi] => $20)
            sio.out(bytes[bi])
          else
            sio.out(".")
      sio.out($d)
      sio.hex(addr+ndx,4)
      sio.str(string(": "))
    sio.hex(byte[addr+ndx],2)
    ndx++
    
  if ndx <> 0
     sio.out(" ")
     repeat bi from 0 to sz-1
       if (bytes[bi] => $20)
         sio.out(bytes[bi])
       else
         sio.out(".")
       
PRI getBreakAddr | sbuf[SBUFLEN+1]
  sio.fgets(@sbuf, SBUFLEN)
  result := sio.strtohex(@sbuf)
  
  if result < jvm.getJvmMain
    sio.Str(String("PC < jem.out main address. Try again. "))
    sio.hex(result,4)
    sio.out(" ")
    sio.hex(jvm.getJvmMain,4)
    result := 0      
  elseif result > jvm.getJvmLength
    sio.Str(String("PC exceeds jem.out size. Try again. "))
    sio.hex(result,4)
    sio.out(" ")
    sio.hex(jvm.getJvmLength,4)
    result := 0      

PRI getBrkPtNum | sbuf[SBUFLEN+1]
  sio.fgets(@sbuf, SBUFLEN)
  result := sio.strtohex(@sbuf)
  if result < 1 or result > getMaxJvmBreakpoints
    sio.Str(String("Breakpoint range 1 to "))
    sio.dec(getMaxJvmBreakpoints)
    sio.Str(string(". Try again. "))
    sio.hex(result,4)
    sio.out(" ")
    sio.hex(jvm.getJvmMain,4)
    result := 0      
    
PUB clearAllJvmBreakpoints | ii 
{{
/**
 * This is a debug control command that clears all breakpoints.
 * The breakpoint queue pointer will be 0 upon return.
 */
__ void clearAllJvmBreakpoints();
}}
  gBreakpointHead := 0
  repeat ii from 0 to BREAKPOINT_SIZE-1
    jvm.setJvmBrkPtByNumber(ii, 0)
    gBreakpointArray[ii] := 0


PUB deleteJvmBrkPtByNum(num) | ii
{{
/**
 * This is a debug control command that lets the user delete a breakpoint by number.
 * @param num is the breakpoint number
 * @returns zero if remove breakpoint failed
 */
__ int deleteJvmBrkPtByNum (int num);
}}
  deleteJvmBreakpoint(gBreakpointArray[num-1])
  result := 0
  
PUB deleteJvmBreakpoint(bppc) | ii, jj 
{{
/**
 * This is a debug control command that lets the user delete a breakpoint.
 * @param bppc is the program counter byte array instruction address.
 * @returns zero if remove breakpoint failed
 */
__ int deleteJvmBreakpoint(bppc);
}}
  result := 0

  ' roll through the queue
  ii := -1  
  repeat while(++ii < BREAKPOINT_SIZE)
    if gBreakpointArray[ii] == bppc
      ' shuffle breakpoints
      repeat while(++ii < BREAKPOINT_SIZE)
        'sio.str(string($d,"bp "))
        'sio.hex(ii, 4)
        gBreakpointArray[ii-1] := gBreakpointArray[ii]
      sio.str(string($d,"deleted bp "))
      sio.hex(ii // BREAKPOINT_SIZE + 1, 2)
      sio.str(string(" at "))
      sio.hex(bppc, 4)
      sio.str(string(" JemPC "))
      sio.hex(bppc-jvm.getJvmStart, 4)
      result := 1
      quit

  repeat ii from 0 to BREAKPOINT_SIZE-1
    jvm.setJvmBrkPtByNumber(ii, gBreakpointArray[ii])


PUB setJvmBreakpoint(bppc) 
{{
/**
 * This is a debug control command that lets the user set one breakpoint.
 * The breakpoint list is an array of BREAKPOINT_SIZE words that behaves like a queue.
 * Only BREAKPOINT_SIZE entries are allowed at one time. The newest entry will replace
 * the oldest entry.
 * @param pc is the program counter byte array instruction address.
 * @returns position of last breakpoint in queue  
 */
__ int setJvmBreakpoint(bppc);
}}
  sio.Str(String("PC $"))
  sio.hex(bppc,4)
  sio.out(" ")
  sio.Str(String($d,"JemPC $"))
  sio.hex(bppc-jvm.getJvmStart,4)
  sio.out(" ")
  result := gBreakpointArray[gBreakpointHead] := bppc
  jvm.setJvmBrkPtByNumber(gBreakpointHead, bppc)
  gBreakpointHead++
  result := gBreakpointHead //= BREAKPOINT_SIZE
  sio.Str(String("Breakpoint "))
  sio.dec(result)
                
PUB getJvmBreakpoint(num) 
{{
/**
 * This is a debug control command that lets the user get the pc of a breakpoint.
 * If the breakpoint input parameter is 0, get the last breakpoint, else get breakpoint specified.
 * The breakpoint number should be <= maximum allowed breakpoints.
 * If the number is greater than maximum, the breakpoint will be the mod of of the number.
 * @param num is the 1-base breakpoint number to get. If num is 0 get the last breakpoint set.
 * @returns pc of selected breakpoint or zero if breakpoint not found. 
 */
__ int getJvmBreakpoints(num);
}}
  if num > 0
    num--
    num //= BREAKPOINT_SIZE
    'result := gBreakpointArray[num]
    result := jvm.getJvmBrkPtByNumber(num)

PUB getMaxJvmBreakpoints 
{{
/**
 * This is a debug control command that tells user maximum number of breakpoints allowed.
 * @returns maximum number of breakpoints. 
 */
__ int getMaxJvmBreakpoints();
}}
  result := jvm.getMaxJvmBreakpoints

PUB skipJvmBreakPointOnce
{{
/**
 * Skip breakpoint once. required to get past breakpoint.
 */
__ void skipJvmBreakPointOnce(void);
}}
  gGetOverBreak := true
  
PUB listJvmBreakpoints | val, ii
{{
/**
 * list all breakpoints
 */
__ void listJvmBreakpoints();
}}
  result := 0
  repeat ii from 1 to BREAKPOINT_SIZE
    val := getJvmBreakpoint(ii)
    if val <> 0
      sio.str(string($d,"Breakpoint "))
      sio.dec(ii)
      sio.str(string(" @PC "))
      sio.hex(val, 4)
      sio.str(string(" JemPC "))
      sio.hex(val-jvm.getJvmStart, 4)
      result++

  if result == 0
    sio.str(string($d,"No Breakpoints Set"))
       


PUB ShowStack | len, ii

  sio.Out( $0D )
  sio.str(string("ST "))
  sio.Hex( jvm.getJvmSP, 4 )
  sio.Out( " " )
  len := (( jvm.getJvmSP - jvm.getJvmSB ) >> jvm#INTWORDS) + 1
  sio.Dec( len)
  sio.out( " " )
  sio.out( $d )
  if len < 0
    len := 15
  if  jvm#INT32
    repeat ii from 0 to len/2 
      if ii and (ii // 16) == 0
        sio.out($d)
      ShowStackLong( ii )
  else
    repeat ii from 0 to len 
      if ii and (ii // 16) == 0
        sio.out($d)
      ShowStackWord( ii )
   
PRI showStackLong( depth )

  sio.out( " " )
  if ( ( jvm.getJvmSP - jvm.getJvmSB ) >> 2 ) => depth
    sio.Hex( long[ jvm.getJvmSP - (depth<<2) ], 8 )
  else
    sio.Str( String("--------") )

PRI showStackWord( depth )

  sio.out( " " )
  if ( ( jvm.getJvmSP - jvm.getJvmSB ) >> 1 ) => depth
    sio.Hex( word[ jvm.getJvmSP - (depth<<1) ], 4 )
  else
    sio.Str( String("----") )

PUB PrintStack(len) | ii
  if  jvm#INT32
    repeat ii from 0 to len/2 
      if ii and (ii // 16) == 0
        sio.out($d)
      ShowStackLong( ii )
  else
    repeat ii from 0 to len 
      if ii and (ii // 16) == 0
        sio.out($d)
      ShowStackWord( ii )
  
PUB ShowHeap | ptr, typ, bytLen, datPtr, nextPtr

  ptr := jvm.getJvmHP
  repeat while ptr < jvm#HEAP_BASE
    nextPtr := jvm.GetHeapInfo( ptr, @typ, @bytLen, 0 )
    ShowHeapEntry( ptr, typ, bytLen )
    ptr := nextPtr
    
PRI ShowHeapEntry( ptr, typ, bytLen )

  sio.Out( $0D )
  sio.Hex( ptr, 4 )
  sio.Out( "-" )
  if ( ChipVer == 1 ) or ( not jvm#BIGARRAYS )
    sio.Hex( ptr + bytLen + ( jvm#HEAP_HEAD_DATA << jvm#HEAP_WORD ) - 1, 4 )
  else
    sio.Hex( ptr + bytLen + ( jvm#HEAP_HEAD_DATA << jvm#HEAP_LONG ) - 1, 4 )
  sio.Out( " " )
  sio.Hex( typ,4)
  if typ < 10
    sio.Out( " " )
  sio.Out( " " )
  sio.Str( jem.TypeNamePtr(typ) )
  sio.Dec( jvm.BitSizePerType(typ) ) 
  sio.Out( "[" )
  sio.dec( bytLen >> jvm.BytesShiftPerType(typ) )
  sio.Out( "]" )  
  sio.Out( " " )
  if ( ChipVer == 1 ) or ( not jvm#BIGARRAYS )
    sio.Dec( jvm#HEAP_HEAD_DATA << jvm#HEAP_WORD )
  else
    sio.Dec( jvm#HEAP_HEAD_DATA << jvm#HEAP_LONG )
  sio.Out( "+" )
  sio.Dec( bytLen )
  sio.Str( String( " Bytes" ) )
  if typ & jem#T_DELETED
    sio.Str( String( " FREE " ) )
    
PUB dumpRegs
  sio.str(string($d,"PC "))
  sio.hex(jvm.getJvmPC, 4)
  sio.str(string(" SP.SB.FP.MP.HP "))
  sio.hex(jvm.getJvmSP, 4)
  sio.Out(".")
  sio.hex(jvm.getJvmSB, 4)
  sio.Out(".")
  sio.hex(jvm.getJvmFP, 4)
  sio.Out(".")
  sio.hex(jvm.getJvmMP, 4)
  sio.Out(".")
  sio.hex(jvm.getJvmHP, 4)
  'sio.Out(".")
  'sio.hex(gClassId, 4)
  sio.Out(" ")
  sio.str(string(" JPC "))
  sio.hex(jvm.getJvmMP-jvm.getJvmStart, 4)
  sio.Out( ":" )
  sio.Hex( jvm.getJvmPC - jvm.getJvmStart, 4 )
{
  sio.Out( " " )
  sio.Dec( pc - mb )
  if (pc - mb ) => 0 and ( pc - mb ) < 10
    sio.Out( " " )
  if ( pc - mb ) < 100
    sio.Out( " " )
}

PUB dumpOpcode(opc, opname)   
  sio.Str(string(" OP "))
  sio.Hex(opc, 2)
  'sio.Str(string(" - "))
  sio.Out(" ")
  sio.Str(opname)
    
PUB JvmSpinDumpBreak(opc, opname) | ch
{{
/**
 * Runs the JVM until a breakpoint is hit or if singlestep is non-zero runs one instruction.
 * @returns 0 if function terminates normally.
 */
__ void JvmSpinDumpBreak(opcode, opname);
}}
  dumpRegs
  dumpOpcode(opc,opname)

DAT
  invalidObjectRef      byte " Invalid Heap or Object reference $", 0
  
PUB ShowObject(object) | ii, ptr, cptr, fptr, type, len, data, clen, flen, cdata, fdata, cid, scid
{{
/**
 * This displays all of objects contents if jazzDebug is enabled.
 * ObjectThisClass, ObjectSuperClass, ObjectFields ... from enum are displayed
 * @param object - object pointer
 */
__ void ShowObject(void* object);
}}
  'if jazzDebug
  'jazzDebug := 0
  if jvm.isHeapPtr(object) == 0
    sio.Str(invalidObjectRef)
    sio.Hex(object, 4 )
    return
    
  jvm.GetHeapInfo(object, @type, @len, @data)  
  sio.Str(string("ShowObject type $"))
  sio.Hex(type, 4 )

  if type <> jem#T_ANY
    sio.Str(invalidObjectRef)
    sio.Hex(object, 4 )
    return  

  ptr := jvm.GetHeapDataValue(data, type, len, jvm#ObjectThisClass)
  cid := ptr
  ptr := jvm.GetHeapDataValue(data, type, len, jvm#ObjectSuperClass)
  scid := ptr

  if cid == 0 or scid == 0
    sio.Str(invalidObjectRef)
    sio.Hex(object, 4 )
    return  

  sio.Str(string(" this $"))
  sio.Hex(cid, 4)

  sio.Str(string(" Super $"))
  sio.Hex(scid, 4 )

{    
  cptr := gConstTable 'GetHeapDataValue(data, type, len, ObjectConTable)
  sio.Str(string(" ConTable $"))
  sio.Hex(cptr, 4 )
  GetHeapInfo(cptr, 0, @clen, @cdata)
  repeat ii from 0 to clen>>1-1
    sio.out(" ")
    sio.hex(GetHeapDataValue(cdata, type, clen, ii),4)
}
  fptr := jvm.GetHeapDataValue(data, type, len, jvm#ObjectFields)
  sio.Str(string(" Fields $"))
  sio.Hex(fptr, 4 )
  jvm.GetHeapInfo(fptr, 0, @flen, @fdata)
  repeat ii from 0 to flen>>1-1
    sio.out(" ")
    sio.hex(jvm.GetHeapDataValue(fdata, type, len, ii),4)
  'jazzDebug := 1

PRI CleanHeap | ptr, typ, bytLen, datPtr, nextPtr, nextTyp, heap

  heap := jvm.getJvmHP  
  ptr := jvm.getJvmHP
  repeat while ptr < jvm#HEAP_BASE
    nextPtr := jvm.GetHeapInfo( ptr, @typ, @bytLen, @datPtr )
    sio.str(string(" heap$"))
    sio.hex(ptr,4)
    sio.str(string(" type$"))
    sio.hex(typ,4)
    if typ & jem#T_DELETED
      if ptr == heap
        heap := nextPtr
      else
        if nextPtr < jvm#HEAP_BASE
          jvm.GetHeapInfo( nextPtr, @nextTyp, 0, 0 )
          if nextTyp & jem#T_DELETED
            ' Combine the two
            nextPtr := jvm.GetHeapInfo( nextPtr, 0, 0, 0 )
            if ( ChipVer == 1 ) or ( not jvm#BIGARRAYS )
              word[ ptr + ( jvm#HEAP_HEAD_BLEN << jvm#HEAP_WORD ) ] := nextPtr - datPtr 
            else
              long[ ptr + ( jvm#HEAP_HEAD_BLEN << jvm#HEAP_LONG ) ] := nextPtr - datPtr  

            sio.str(string(" clean$"))
            sio.hex(ptr,4)
            sio.str(string(" -> $"))
            sio.hex(nextPtr,4)

            nextPtr := ptr
    sio.out($d)
    ptr := nextPtr
    